/* debug.c
 * functions to implement the debugger
 * Author: John Lenz <jelenz@wisc.edu>
 * (C) Copyright 2003 John Lenz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <inttypes.h>

static const int MAX_LINE_SIZE = 256;

/* variables from main.c */
extern int16_t pc;
extern int run;
extern int16_t mem[];
extern int16_t reg[];

/* these arrays are zero based so line numbers start counting from 0
 * when printing out the line number, we need to add 1
 * so that the first line is line # 1. */
char **lines = NULL; /* lines from the source file */
unsigned int *pc_to_line = NULL; /* mapping from PC to source line # */

/* breakpoints stores the line number number to break (where the first line is # 1) 
 * so when comparing we need to subtract 1 so that the first line is # 0
 * pc_to_line[pc] == breakpoints[1] - 1
 * a zero in breakpoints means that the breakpoint was deleted */
unsigned int breakpoints[100];
int num_breakpoints = 0;

/* holds the pc of the previous instruction (used for single stepping instructions) */
int32_t prev_instr = -1;
      
/* ----------------------------------------------------------------------------
 * void debug_load_data(help_name, asm_name)
 *  load asm_name into char **lines and help_name into unsigned int *pc_to_line
 * --------------------------------------------------------------------------*/
void debug_load_data(char *help_name, char *asm_name) {
  int nlines = 0;
  char c;
  int temp;
  FILE *asm_file, *help_file;
  char str[MAX_LINE_SIZE];
  int addr, line;
  
  asm_file = fopen(asm_name, "r");
  if (asm_file == NULL) {
    fprintf(stderr, "Unable to open %s\n", asm_name);
    exit(1);
  }
  
  while ((c = getc(asm_file)) != EOF) {
    if (c == '\n') nlines++;
  }

  fseek(asm_file, 0, SEEK_SET);

  lines = (char **) calloc(nlines, sizeof(char*));
  pc_to_line = (unsigned int *) calloc(nlines, sizeof(unsigned int));
  for (nlines = 0; ; nlines++) {
    if (fgets(str, MAX_LINE_SIZE, asm_file) == NULL) break;
    lines[nlines] = (char *) malloc(strlen(str) + 1);
    strcpy(lines[nlines], str);
  }
  fclose(asm_file);

  help_file = fopen(help_name, "r");
  if (help_file == NULL) {
    fprintf(stderr, "Unable to open %s\n", help_name);
    exit(1);
  }

  /* skip the Variable Addresses section */
  while (1) {
    if (fgets(str, MAX_LINE_SIZE, help_file) == NULL) goto close_file;
    if (strcmp(str, "Instruction Addresses\n") == 0) break;
  }

  /* read in the Instruction Addresses section */
  for (nlines = 0; fscanf(help_file, "%x %i", &addr, &line) != EOF; nlines++) {
    pc_to_line[addr] = line - 1;
  } 

close_file:
  fclose(help_file);

  /* break on the first instruction */
  prev_instr = 0;
}


/* ----------------------------------------------------------------------------
 * void debug_check_break()
 *     called between every instruction 
 * --------------------------------------------------------------------------*/
void debug_check_break() {
  int i;
  int cont = 1;

  for (i = 0; i < num_breakpoints; i++) {
    if (breakpoints[i] - 1 == pc_to_line[pc] && breakpoints[i] > 0) break;
  }
  if (prev_instr == -1) { /* If single stepping is off */
    if (i >= num_breakpoints || (pc > 0 && pc_to_line[pc - 1] == pc_to_line[pc])) return;
    printf("Breakpoint %i\n", i + 1);
  } else if (pc_to_line != NULL) {
    /* if more than one instruction was generated for a line
     * then skip the second instruction (for example, the li pseudo-instruction) */
    if (pc_to_line[prev_instr] == pc_to_line[pc]) return;
  }

  if (pc_to_line != NULL && lines != NULL) {
    if (pc_to_line[pc] != 0 && lines[pc_to_line[pc]] != NULL)
      printf("%i: %s", pc_to_line[pc] + 1, lines[pc_to_line[pc]]);
    else
      printf("No source for instruction at address %04hx: %04hx\n", pc, mem[pc]);
  } else {
    printf("No source for instruction at address %04hx: %04hx\n", pc, mem[pc]);
  }
 
  /* prompt for command */
  while (cont) {
    int temp;
    char *cmd = NULL;
    char input[MAX_LINE_SIZE];
    printf("> ");
    if (fgets(input, MAX_LINE_SIZE, stdin) == NULL) {
      prev_instr = -1;
      return;
    }
  
    cmd = (char *) strtok(input, " ");
    
      /* go to the next instruction */
    if (strcmp(cmd, "n\n") == 0) {
      cont = 0;
      prev_instr = pc;

      /* print out some value */
    } else if (strcmp(cmd, "p") == 0) {
      char *arg = (char *) strtok(NULL, " ");
      if (arg[0] == '$') {
        int r = -1;
        if (strncmp(arg, "$ra", 3) == 0)
          r = 7;
        else if (strncmp(arg, "$sp", 3) == 0)
          r = 0;
        else if (strncmp(arg, "$fp", 3) == 0)
          r = 1;
        else if (strncmp(arg, "$t0", 3) == 0)
          r = 2;
        else if (strncmp(arg, "$t1", 3) == 0)
          r = 3;
        else if (strncmp(arg, "$v0", 3) == 0)
          r = 4;
        else if (strncmp(arg, "$v1", 3) == 0)
          r = 5;
        else {
          if (sscanf(arg, "$%i", &r) == 0) r = -1;
        }
        if (r >= 0)
          printf("$%i = %04hx, %hi\n", r, reg[r], reg[r]);
      } else if (strncmp(arg, "pc", 2) == 0) {
        printf("pc = %04hx\n", pc);
      } else {
        int m;
        if (sscanf(arg, "%i", &m) > 0)
          printf("mem[%i] = %04hx, %hi\n", m, mem[m], mem[m]);
      }

      /* quit the program */
    } else if (strcmp(cmd, "quit\n") == 0) {
      cont = 0;
      run = 0;

      /* continue executing */
    } else if (strcmp(cmd, "c\n") == 0) {
      cont = 0;
      prev_instr = -1;

      /* Add a breakpoint */
    } else if (strcmp(cmd, "b") == 0) {
      if (num_breakpoints < 100) {
        char *arg = (char *) strtok(NULL, " ");
        int l;
        if (sscanf(arg, "%i", &l) > 0) {
          breakpoints[num_breakpoints++] = l;
          printf("Added breakpoint %i at line %i\n", num_breakpoints, l);
        }
      } 

      /* Delete a breakpoint */
    } else if (strcmp(cmd, "d") == 0) {
      char *arg = (char *) strtok(NULL, " ");
      int l;
      if (sscanf(arg, "%i", &l) > 0) {
        if (l > 0 && l < 101) {
          breakpoints[l - 1] = 0;
          printf("Deleted breakpoint %i\n", l);
        }
      }

      /* execute the n instruction */
    } else if (cmd[0] == '\n' && prev_instr >= 0) {
      cont = 0;
      prev_instr = pc;
      
    } else {
      printf("Bad command\n");
    }
  }
}
