/* main.c
 *   main file for WISC emulator
 * Author: John Lenz <jelenz@wisc.ed>
 * (C) Copyright 2003 John Lenz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <inttypes.h>
#include "opcode.h"

/* machine variables */
#define MEM_SIZE 65536
#define REG_SIZE 8
int16_t reg[REG_SIZE];
int16_t mem[MEM_SIZE];
int16_t pc;
int argnum;

void usage(char *name) {
  fprintf(stderr, "%s prog.mem [prog.help prog.s]\n", name);
  exit(1);
}

/* debug functions (in debug.c) */
void debug_load_data(char *, char *);
void debug_check_break();

/* global flags */
int run = 1;

/* read in arguments */
void read_args(int argc, char **argv) {
  FILE *prog;

  if (argc < 2) usage(argv[0]);

  argnum = 1;
  /*if (strcmp(argv[1], "-w") == 0) {
    trap = 0;
    argnum = 2;
    if (argc < 3) usage(argv[0]);
  }*/

  prog = fopen(argv[argnum], "r");
  if (prog == NULL) {
    fprintf(stderr, "Unable to open %s\n", argv[argnum]);
    exit(1);
  }

  if (argc > 2 + argnum) {
    debug_load_data(argv[argnum + 1], argv[argnum + 2]);
  } else if (argc > 1 + argnum) {
    usage(argv[0]);
  }

  int32_t addr, val;
  while (fscanf(prog, "%x %x", &addr, &val) != EOF) {
    mem[addr] = val;
  }
  fclose(prog);
}

/* dump memory to file */
void dump_mem(char *name) {
  char out[strlen(name) + 5];
  memcpy(out, name, strlen(name));
  strcpy(out + strlen(name), ".out");

  FILE *outf = fopen(out, "w");
  if (outf == NULL) {
    fprintf(stderr, "Unable to open %s", out);
    exit(1);
  }

  fprintf(outf, "Addr Data Data in Decimal\n");

  int32_t i;
  for (i = 0; i < MEM_SIZE; i++) {
    if (mem[i] != 0)
      fprintf(outf, "%04hx %04hx %hi\n", i, mem[i], mem[i]);
  }

  fclose(outf);
}

#define sign_extend(num, size, mask) \
  (((1 << (size - 1)) & num) ? (mask | num) : num)

/* this is machine dependent... C does not define what type of shift is used
 * on some machines (the one I am using at least) follows the rules that
 * if the operand is an unsigned type a logical shift is used
 * if the operand is a signed type then an arithmatic shift is used */
#define shift_l(num, size) ((uint16_t)num >> size)
#define shift_a(num, size) (num >> size)

#define rotate(num, size) \
  ((num << 16 - (size % 16)) | shift_l(num, (size % 16)))


int main(int argc, char **argv) {
  memset(mem, 0, sizeof(int16_t) * MEM_SIZE);
  memset(reg, 0, sizeof(int16_t) * REG_SIZE);
  pc = 0;

  read_args(argc, argv);

  int16_t instr, opcode, rd, rs, rt;
  int16_t i5_zero, i5_sign, i8, i11, func;

  while (run) {
    debug_check_break();
    instr = mem[pc];
    pc += 1;
    
    opcode = (0xF800 & instr) >> 11;
    rd = (0x0700 & instr) >> 8;
    rs = (0x00E0 & instr) >> 5;
    rt = (0x001C & instr) >> 2;
    func = (0x0003 & instr);
    i5_zero = (0x001F & instr);
    i5_sign = sign_extend(i5_zero, 5, 0xFFE0);
    i8 = (0x00FF & instr);
    i8 = sign_extend(i8, 8, 0xFF00);
    i11 = (0x07FF & instr);
    i11 = sign_extend(i11, 11, 0xF800);

    switch (opcode) {
      case HALT:
        run = 0;
        break;

      case WRITE:
        switch (func) {
          case WRINT:
            printf("%hi", reg[rs]);
            break;
          case WRCH:
            printf("%c", reg[rs]);
            break;
        }
        break;

      case ADDI:
        reg[rd] = reg[rs] + i5_sign;
        break;
      case ANDI:
        reg[rd] = reg[rs] & i5_zero;
        break;
      case ORI:
        reg[rd] = reg[rs] | i5_zero;
        break;
      case XORI:
        reg[rd] = reg[rs] ^ i5_zero;
        break;
      case SUBI:
        reg[rd] = reg[rs] - i5_sign;
        break;
      case RORI:
        reg[rd] = rotate(reg[rs], i5_zero);
        break;
      case SRLI:
        reg[rd] = shift_l(reg[rs], i5_zero);
        break;
      case SRAI:
        reg[rd] = shift_a(reg[rs], i5_zero);
        break;

      case SUB_NEG:
        switch (func) {
          case SUB:
            reg[rd] = reg[rs] - reg[rt];
            break;
          case NEG:
            reg[rd] = -reg[rt];
            break;
        }
        break;

      case NOT:
        reg[rd] = ~ reg[rt];
        break;

      case ALU:
        switch (func) {
          case ADD:
            reg[rd] = reg[rs] + reg[rt];
            break;
          case AND:
            reg[rd] = reg[rs] & reg[rt];
            break;
          case OR:
            reg[rd] = reg[rs] | reg[rt];
            break;
          case XOR:
            reg[rd] = reg[rs] ^ reg[rt];
            break;
        }
        break;

      case SHIFT:
        switch (func) {
          case ROR:
            reg[rd] = rotate(reg[rs], reg[rt]);
            break;
          case SRL:
            reg[rd] = shift_l(reg[rs], reg[rt]);
            break;
          case SRA:
            reg[rd] = shift_a(reg[rs], reg[rt]);
            break;
        }
        break;

      case LDI:
        reg[rd] = mem[(uint16_t)(reg[rs] + i5_sign)];
        break;
      case STI:
        mem[(uint16_t)(reg[rs] + i5_sign)] = reg[rd];
        break;
      case LD:
        reg[rd] = mem[(uint16_t)(reg[rs] + reg[rt])];
        break;
      case ST:
        mem[(uint16_t)(reg[rs] + reg[rt])] = reg[rd];
        break;
      case LUI:
        reg[rd] = (0x00FF & instr) << 8;
        break;
      case LLI:
        reg[rd] = (0xFF00 & reg[rd]) | (0x00FF & instr);
        break;
      case LDPLUS:
        reg[rd] = mem[(uint16_t)(reg[rs] + i5_sign)];
        reg[rs] = reg[rs] + 1;
        break;

      case BEQZ:
        if (reg[rd] == 0) pc += i8;
        break;
      case BNEZ:
        if (reg[rd] != 0) pc += i8;
        break;
      case BLEZ:
        if (reg[rd] <= 0) pc += i8;
        break;
      case BLTZ:
        if (reg[rd] < 0) pc += i8;
        break;
      case BDNE:
        if (reg[rd] != 0) {
          pc += i8;
          reg[rd] = reg[rd] - 1;
        }
        break;

      case RET:
        pc = reg[7];
        break;
      case JUMP:
        pc += i11;
        break;
      case JAL:
        reg[7] = pc;
        pc += i11;
        break;
      case JR:
        pc = reg[rd];
        break;
      case JALR:
        reg[7] = pc;
        pc = reg[rd];
        break;

    }
  }

  dump_mem(argv[argnum]);
  return 0;
}
